# $Id: dr_choose-server.pl $
#
# dr_choose-server.pl is plugin to implementing dead-realm feature for
# Radiator as replacement for dead-host feature which is not suitable
# for proxy networks such as eduroam.
#
# Author: Jan Tomek <jan.tomasek@cesnet.cz>, <jan@tomasek.cz>
# Copyright (c) 2006-8, CESNET, z.s.p.o.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# See README and COPYING for more details.

package DeadRealm;

# Maximum number of dead realm records which will be kept in cache for
# each defined host. Actualy number of kept records will vary from
# max*0.9 to max.
#
# For 10000 records is needed 11384-9508=1876kB memory (length of
# realm was "test09999.orgA.etest.cesnet.cz").
#
my $max_dead_realm_records = 1000;

sub getRealm {
  my $username = shift;

  return $1 if ($username =~ /@(.+)$/);
  return undef;
}

sub replyHook {
  my $username = ${$_[3]}->get_attr('User-Name');
  my $hostname = ${$_[3]}->{ThisHost}->{Name};
  my $realm = getRealm($username) || 'unknown-realm';

  if (defined($main::deadRealms->{$hostname}->{$realm})) {
    # Remove dead mark
    undef $main::deadRealms->{$hostname}->{$realm};
    &main::log($main::LOG_INFO,
	       "dead-realm: Clearing dead mark for \@$realm on host $hostname.");
  };

  return;
};

sub noReplyHook {
  my $username = ${$_[1]}->get_attr('User-Name');
  my $hostname = ${$_[1]}->{ThisHost}->{Name};
  my $realm = getRealm($username) || 'unknown-realm';

  my $dead_realm_records = scalar(keys %{$main::deadRealms->{$hostname}});
  if ($dead_realm_records > $max_dead_realm_records) {
    # Number of dead realm record exceeded
    # $max_dead_realm_records. Someone is propably trying to fill our
    # memory by trash. We have to delete oldest records.
    foreach my $realm (sort {$main::deadRealms->{$hostname}->{$a} <=> $main::deadRealms->{$hostname}->{$b}} keys %{$main::deadRealms->{$hostname}}) {
      &main::log($main::LOG_INFO,
		 "dead-realm: removing realm \@$realm from host $hostname cache ($dead_realm_records).");
      delete $main::deadRealms->{$hostname}->{$realm};
      $dead_realm_records--;
      last if ($dead_realm_records < ($max_dead_realm_records*0.9));
    };
  };

  # Mark this host dead for this realm
  $main::deadRealms->{$hostname}->{$realm} = time;
  &main::log($main::LOG_INFO,
	     "dead-realm: Marking realm \@$realm dead on host $hostname cache ($dead_realm_records).");
  return;
};

sub chooseServer {
  my @args = @_;

  # Get parameters from radiator config
  my $timeout = &main::getVariable('dr_mark_timeout') || 60*60;
  my $identifier = $args[0]->{Handler}->{Identifier};
  my $servers = &main::getVariable('dr_'.$identifier.'_server_list');
  my @servers = split(' *, *', $servers);
  unless (@servers) {
    &main::log($main::LOG_ERROR,
	       "dead-realm: There are not defined servers for Handler Identifier=$identifier.");
    return undef;
  };

  my $username = $args[0]->get_attr('User-Name');
  my $realm = getRealm($username) || 'unknown-realm';

  my %servers;

  # At first we try to find server which is has not marked realm we
  # are processing now as dead, or server which have that realm marked
  # as for too long.
  #
  # If you wish to stay with one server until it goes down set set
  # $dr_mark_timeout unbelivable big value. Such as one year
  # 365*24*60*60.
  foreach my $server (@servers) {
    if (defined($main::deadRealms->{$server}->{$realm})) {
      # Check if this realm is not makred as dead for too long.
      if (time - $main::deadRealms->{$server}->{$realm} > $timeout) {
	undef $main::deadRealms->{$server}->{$realm};
	&main::log($main::LOG_DEBUG,
		   "dead-realm: Clearing too old dead mark for \@$realm on host $server.");
      } else {
	# Queue this server for later if we will not find any better
	$servers{$server} = $main::deadRealms->{$server}->{$realm};
	next;
      };
    };

    my $auth = Radius::AuthGeneric::find($server);
    if ($auth) {
      return $auth->handle_request(@args);
    } else {
      &main::log($main::LOG_ERROR, "dead-realm: Can't locate AuthBy handler $server");
      return undef;
    };
  };

  # If we failed to find server which has not dead mark for realm we
  # are processing, we do not give up! We will try to send our packet
  # to server with oldest dead-realm mark. Chosing server with oldest
  # dead-realm mark cause that, we will cycle throught list untill
  # some of them will become alive.
  #
  # If server is dead, than it's dead-realm mark is updated by
  # NoReplyHook in it's Auth block.
  #
  # If server is alive, than it's dead-realm mark is removed at moment
  # we receive some response from from it. That is done by ReplyHook
  # in it's Auth block.
  foreach my $server (sort {$servers{$a} <=> $servers{$b}} keys %servers) {
    my $auth = Radius::AuthGeneric::find($server);
    return $auth->handle_request(@args) if ($auth);
  };

  # We are loosers there is not any valid AuthBy ... That is probably
  # missconfiguration.
  &main::log($main::LOG_ERROR, "dead-realm: Can't locate ANY defined AuthBy");
  return undef;
};

1;
